/* cal_cons.c - routines for printing translated alignments for
   fasta, ssearch, ggsearch, glsearch  */

/* copyright (c) 1998, 1999, 2007 by William R. Pearson and the University of Virginia */

/* $Id: cal_cons.c 1172 2013-06-06 14:45:20Z wrp $ */
/* $Revision: 1172 $  */

/* removed from dropgsw2.c, dropnfa.c April, 2007 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "defs.h"
#include "param.h"

#if defined(FASTA) || defined(TFASTA)
#include "dropnfa.h"
#endif

#if defined(SSEARCH) || defined(OSEARCH)
#include "dropgsw2.h"
#endif

#ifdef LALIGN
#include "dropgsw2.h"
#endif

#include "a_mark.h"

static void update_code(char *al_str, int al_str_max, int op, int op_cnt, int show_code, int *start_code);
extern void aancpy(char *to, char *from, int count, struct pstruct *ppst);
extern void *init_stack(int, int);
extern void push_stack(void *, void *);
extern void *pop_stack(void *);
extern void *free_stack(void *);

/* returns M_NEG, M_ZERO, M_POS, M_IDENT, M_DEL (a_mark.h)
   updates *aln->nsim, npos, nident, nmismatch */
extern int
align_type(int score, char sp0, char sp1, int nt_align, struct a_struct *aln, int pam_x_id_sim);

extern int
next_annot_match(int *itmp, int *pam2aa0v, long ip, long ia, 
		 char *sp1, char *sp1a, const unsigned char *sq,
		 int i_annot, int n_annot, struct annot_entry **s_annot_arr, char **ann_comment,
		 void *annot_stack, int *have_push_features, int *v_delta,
		 struct annot_entry **region_p, struct annot_entry *tmp_region_p, int init_score);

extern void
comment_var(long i0, char sp0, long i1, char sp1, char o_sp1, char sim_char,
	    const char *ann_comment, char *annot_var_s, int annot_var_n,
	    int target, int d_type);

void
display_push_features(void *annot_stack, char *annot_var_s, int n_annot_var_s, 
		      long i0_pos, char sp0, long i1_pos, char sp1, char sym, 
		      struct annot_entry **region0_p,
		      struct annot_entry **region1_p,
		      struct rstruct *rst, int n0, int n1,
		      void *pstat_void, int d_type);

#define DP_FULL_FMT 1	/* Region: score: bits: id: ... */

extern int seq_pos(int pos, int rev, int off);

int calc_cons_a(const unsigned char *aa0, int n0,
		const unsigned char *aa1, int n1,
		int *nc,
		struct a_struct *aln,
		struct a_res_str *a_res,
		struct pstruct *ppst,
		char *seqc0, char *seqc1, char *seqca, int *seqc_score,
		const unsigned char *ann_arr,
		const unsigned char *aa0a, const struct annot_str *annot0_p, char *seqc0a,
		const unsigned char *aa1a, const struct annot_str *annot1_p, char *seqc1a,
		int *score_delta, 
		char *annot_var_s, int n_annot_var_s,
		struct f_struct *f_str,
		void *pstat_void)
{
  int i0, i1, nn1;
  int op, lenc, nd, ns, itmp, p_match;
  const unsigned char *aa1p;
  char *sp0, *sp0a, *sp1, *sp1a, *spa, t_spa;
  int *i_spa;
  const unsigned char *sq;
  int *rp;
  int smins, mins, ntmp;
  int have_ann = 0;
  void *annot_stack;

  /* variables for variant changes */
  char *sim_sym = aln_map_sym[5];
  struct annot_entry **s_annot0_arr_p, *region0_p, tmp_annot0;
  struct annot_entry **s_annot1_arr_p, *region1_p, tmp_annot1;
  char *ann_comment;
  int i0_annot, i1_annot;	/* i0_annot, i1_annot, count through
				   the list of annotations */
  int v_delta, v_tmp;
  int have_push_features;

  /* variables for handling coordinate offsets */
  long q_offset, l_offset;

  *score_delta = 0;
  annot_var_s[0] = '\0';
  have_ann = (seqc0a != NULL); 

  if (ppst->ext_sq_set) {
    sq = ppst->sqx;
  }
  else {
    sq = ppst->sq;
  }

#ifndef TFASTA
  aa1p = aa1;
  nn1 = n1;
#else
  aa1p = f_str->aa1x;
  nn1 = f_str->n10;
#endif

  aln->amin0 = a_res->min0;
  aln->amax0 = a_res->max0;
  aln->amin1 = a_res->min1;
  aln->amax1 = a_res->max1;
  aln->calc_last_set = 1;

  q_offset = aln->q_offset;
  l_offset = aln->l_offset;

#ifndef LCAL_CONS
  /* will we show all the start ?*/
  if (min(a_res->min0,a_res->min1)<aln->llen || aln->showall==1)
    if (a_res->min0>=a_res->min1) {              /* aa0 extends more to left */
      smins=0;
      if (aln->showall==1) mins = a_res->min0;
      else mins = min(a_res->min0,aln->llcntx);
      aancpy(seqc0,(char *)aa0+a_res->min0-mins,mins,ppst);
      aln->smin0 = a_res->min0-mins;
      if ((mins-a_res->min1)>0) {
	memset(seqc1,' ',mins-a_res->min1);
	aancpy(seqc1+mins-a_res->min1,(char *)aa1p,a_res->min1,ppst);
	aln->smin1 = 0;
      }
      else {
	aancpy(seqc1,(char *)aa1p+a_res->min1-mins,mins,ppst);
	aln->smin1 = a_res->min1-mins;
      }
    }
    else {
      smins=0;
      if (aln->showall == 1) mins=a_res->min1;
      else mins = min(a_res->min1,aln->llcntx);
      aancpy(seqc1,(char *)(aa1p+a_res->min1-mins),mins,ppst);
      aln->smin1 = a_res->min1-mins;
      if ((mins-a_res->min0)>0) {
	memset(seqc0,' ',mins-a_res->min0);
	aancpy(seqc0+mins-a_res->min0,(char *)aa0,a_res->min0,ppst);
	aln->smin0 = 0;
      }
      else {
	aancpy(seqc0,(char *)aa0+a_res->min0-mins,mins,ppst);
	aln->smin0 = a_res->min0-mins;
      }
    }
  else {
    mins= min(aln->llcntx,min(a_res->min0,a_res->min1));
    smins=mins;
    aln->smin0=a_res->min0 - smins;
    aln->smin1=a_res->min1 - smins;
    aancpy(seqc0,(char *)aa0+a_res->min0-mins,mins,ppst);
    aancpy(seqc1,(char *)aa1p+a_res->min1-mins,mins,ppst);
  }
  /* set the alignment code to zero for context */
  memset(seqca,0,mins);
  if (have_ann) {
    memset(seqc0a,' ',mins);
    memset(seqc1a,' ',mins);
  }
#else		/* no flanking context */
  smins = mins = 0;
  aln->smin0=a_res->min0;
  aln->smin1=a_res->min1;
#endif

  /* now get the middle */

  spa = seqca+mins;
  if (seqc_score) i_spa = seqc_score+mins;
  sp0 = seqc0+mins;
  sp1 = seqc1+mins;

  if (have_ann) {
    sp0a = seqc0a+mins;
    sp1a = seqc1a+mins;
  }

  rp = a_res->res;
  lenc = aln->nident = aln->nmismatch =
    aln->npos = aln->nsim = aln->ngap_q = aln->ngap_l = aln->nfs = op = 0;
  p_match = 1;
  i0 = a_res->min0;
  i1 = a_res->min1;

  v_delta = 0;
  i0_annot = i1_annot = 0;
  annot_stack = NULL;
  region0_p = region1_p = NULL;
  s_annot0_arr_p = s_annot1_arr_p = NULL;
  have_push_features=0;
  if (have_ann) {
    if ((annot1_p && annot1_p->n_annot>0) || (annot0_p && annot0_p->n_annot > 0)) {annot_stack = init_stack(64,64);}
    if (annot1_p && annot1_p->n_annot > 0) {
      s_annot1_arr_p = annot1_p->s_annot_arr_p;
      while (i1_annot < annot1_p->n_annot && s_annot1_arr_p[i1_annot]->pos < i1 + l_offset) {
	if (s_annot1_arr_p[i1_annot]->label == '[') {
	  memcpy(&tmp_annot1,s_annot1_arr_p[i1_annot], sizeof(struct annot_entry));
	  tmp_annot1.a_pos = a_res->min0 + q_offset;
	  tmp_annot1.pos = a_res->min1 + l_offset;
	  region1_p = &tmp_annot1;
	  region1_p->score = region1_p->n_aln = region1_p->n_ident = 0;
	}
	else if (s_annot1_arr_p[i1_annot]->label == ']') {
	  region1_p = NULL;
	}
	i1_annot++;
      }
    }

    if (annot0_p && annot0_p->n_annot>0) {
      s_annot0_arr_p = annot0_p->s_annot_arr_p;

      while (i0_annot < annot0_p->n_annot && s_annot0_arr_p[i0_annot]->pos < i0 + q_offset) {
	if (s_annot0_arr_p[i0_annot]->label == '[') {
	  memcpy(&tmp_annot0,s_annot0_arr_p[i0_annot], sizeof(struct annot_entry));
	  tmp_annot0.pos = a_res->min0 + q_offset;
	  tmp_annot0.a_pos = a_res->min1 + l_offset;
	  region0_p = &tmp_annot0;
	  region0_p->score = region0_p->n_aln = region0_p->n_ident = 0;
	}
	else if (s_annot0_arr_p[i0_annot]->label == ']') {
	  region0_p = NULL;
	}
	i0_annot++;
      }
    }
  }

  while (i0 < a_res->max0 || i1 < a_res->max1) {
    /* match/mismatch (aligned residues */
    /* here, op is the "current" encoding, and *rp is the next one */
    if (op == 0 && *rp == 0) {
      op = *rp++;
      lenc++;

      itmp=ppst->pam2[0][aa0[i0]][aa1p[i1]];
      *sp0 = sq[aa0[i0]];
      *sp1 = sq[aa1p[i1]];

      if (have_ann) {
	have_push_features = 0;
	*sp0a = *sp1a = ' ';
	if (aa0a) {*sp0a = ann_arr[aa0a[i0]];}
	if (aa1a) {*sp1a = ann_arr[aa1a[i1]];}
	if (s_annot1_arr_p) {
	  if (i1+l_offset == s_annot1_arr_p[i1_annot]->pos) {
	    i1_annot = next_annot_match(&itmp, ppst->pam2[0][aa0[i0]], l_offset+seq_pos(i1,aln->llrev,0),
					q_offset + seq_pos(i0,aln->qlrev,0), sp1, sp1a, sq, 
					i1_annot, annot1_p->n_annot, s_annot1_arr_p,
					&ann_comment, annot_stack, &have_push_features, &v_delta,
					&region1_p, &tmp_annot1,0);

	    /* must be out of the loop to capture the last value */
	    if (sq[aa1p[i1]] != *sp1) {
	      t_spa = align_type(itmp, *sp0, *sp1, ppst->nt_align, NULL, ppst->pam_x_id_sim);

	      comment_var(q_offset+seq_pos(i0,aln->qlrev,0), *sp0,
			  l_offset+seq_pos(i1,aln->llrev,0), *sp1,
			  sq[aa1p[i1]], sim_sym[t_spa], ann_comment,
			  annot_var_s, n_annot_var_s, 1, 1);
	    }
	  }
	  if (region1_p) {
	    region1_p->score += itmp;
	  }
	}

	if (s_annot0_arr_p) {
	  if (i0 + q_offset == s_annot0_arr_p[i0_annot]->pos) {
	    i0_annot = next_annot_match(&itmp, ppst->pam2[0][aa1p[i1]], q_offset+seq_pos(i0,aln->qlrev,0),
					l_offset+seq_pos(i1,aln->llrev,0), sp0, sp0a, sq, 
					i0_annot, annot0_p->n_annot, s_annot0_arr_p,
					&ann_comment, annot_stack, &have_push_features, &v_delta,
					&region0_p, &tmp_annot0, 0);

	    /* must be out of the loop to capture the last value */
	    if (sq[aa0[i0]] != *sp0) {
	      t_spa = align_type(itmp, *sp0, *sp1, ppst->nt_align, NULL, ppst->pam_x_id_sim);

	      comment_var(q_offset+seq_pos(i0,aln->qlrev,0), *sp0,
			  l_offset+seq_pos(i1,aln->llrev,0), *sp1,
			  sq[aa0[i0]], sim_sym[t_spa], ann_comment,
			  annot_var_s, n_annot_var_s, 0, 1);
	    }
	  }
	  if (region0_p) { 
	    region0_p->score += itmp;
	  }
	}
	sp0a++; sp1a++;
      }

      *spa = align_type(itmp, *sp0, *sp1, ppst->nt_align, aln, ppst->pam_x_id_sim);

      if (region1_p) {
	region1_p->n_aln++;
	if (*spa == M_IDENT) {region1_p->n_ident++;}
      }

      if (region0_p) {
	region0_p->n_aln++;
	if (*spa == M_IDENT) {region0_p->n_ident++;}
      }

      /* now we have done all the ?modified identity checks, display
	 potential site annotations */
      if (have_ann && have_push_features) {
	display_push_features(annot_stack, annot_var_s, n_annot_var_s,
			      q_offset+seq_pos(i0,aln->qlrev,0), *sp0,
			      l_offset+seq_pos(i1,aln->llrev,0), *sp1,
			      sim_sym[*spa], &region0_p, &region1_p,
			      &(a_res->rst), n0, n1, pstat_void, DP_FULL_FMT);
	have_push_features = 0;
      }

      if (seqc_score) *i_spa++ = itmp;
      i0++; i1++;
      sp0++; sp1++; spa++;
    }
    else {	/* indel */
      if (op==0) {
	op = *rp++;
	if (seqc_score) *i_spa = ppst->gdelval;
	if (annot1_p && region1_p) { region1_p->score +=  ppst->gdelval;}
	if (annot0_p && region0_p) { region0_p->score +=  ppst->gdelval;}
      }
      if (seqc_score) *i_spa++ += ppst->ggapval;
      if (s_annot1_arr_p && region1_p) { region1_p->score +=  ppst->ggapval; region1_p->n_aln++;}
      if (s_annot0_arr_p && region0_p) { region0_p->score +=  ppst->ggapval; region0_p->n_aln++;}

      if (op>0) {	/* insertion in aa0 */
	*sp1 = sq[aa1p[i1]];
	*sp0 = '-';
	*spa++ = M_DEL;
	if (have_ann) {
	  have_push_features = 0;
	  *sp0a++ = ' ';
	  if (aa1a) *sp1a = ann_arr[aa1a[i1]];
	  else *sp1a = ' ';
	  if (s_annot1_arr_p) {
	    if (i1+l_offset == s_annot1_arr_p[i1_annot]->pos) {
	      i1_annot = next_annot_match(&itmp, ppst->pam2[0][aa0[i0]], l_offset+seq_pos(i1,aln->llrev,0),
					  q_offset+seq_pos(i0,aln->qlrev,0), sp1, sp1a, sq, 
					  i1_annot, annot1_p->n_annot, s_annot1_arr_p,
					  &ann_comment, annot_stack, &have_push_features, &v_delta,
					  &region1_p, &tmp_annot1, ppst->ggapval+ppst->gdelval);
	    }
	  }

	  if (have_push_features) {
	    display_push_features(annot_stack, annot_var_s, n_annot_var_s,
				  q_offset+seq_pos(i0,aln->qlrev,0), *sp0,
				  l_offset+seq_pos(i1,aln->llrev,0), *sp1,
				  sim_sym[*spa], &region0_p, &region1_p,
				  &(a_res->rst), n0, n1, pstat_void, DP_FULL_FMT);
	    have_push_features = 0;
	  }
	  sp1a++;
	}

	sp0++;
	sp1++;
	i1++;
	op--;
	lenc++;
	aln->ngap_q++;
      }
      else {		/* insertion in aa1 */
	*sp1 = '-';
	*spa++ = M_DEL;
	*sp0 = sq[aa0[i0]];
	if (have_ann) {
	  have_push_features = 0;
	  *sp1a++ = ' ';
	  if (aa0a) *sp0a = ann_arr[aa0a[i0]];
	  else *sp0a = ' ';
	  if (s_annot0_arr_p) {
	    if (i0+q_offset == s_annot0_arr_p[i0_annot]->pos) {
	      i0_annot = next_annot_match(&itmp, ppst->pam2[0][aa0[i0]], q_offset+seq_pos(i0,aln->qlrev,0),
					  l_offset+seq_pos(i1,aln->llrev,0), sp0, sp0a, sq, 
					  i0_annot, annot0_p->n_annot, s_annot0_arr_p,
					  &ann_comment, annot_stack, &have_push_features, &v_delta,
					  &region0_p, &tmp_annot0, ppst->ggapval+ppst->gdelval);
	    }
	  }

	  if (have_push_features) {
	    display_push_features(annot_stack, annot_var_s, n_annot_var_s,
				  q_offset+seq_pos(i0,aln->qlrev,0), *sp0,
				  l_offset+seq_pos(i1,aln->llrev,0), *sp1,
				  sim_sym[*spa], &region0_p, &region1_p,
				  &(a_res->rst), n0, n1, pstat_void, DP_FULL_FMT);
	    have_push_features = 0;
	  }
	  sp0a++;
	}

	i0++;
	sp0++;
	sp1++;
	op++;
	lenc++;
	aln->ngap_l++;
      }
    }
  }

  *score_delta = v_delta;

  *nc = lenc;
  if (have_ann) {
    *sp0a = *sp1a = '\0';
    have_push_features = 0;
    if (annot1_p) {
      /* also check for regions after alignment */
      while (i1_annot < annot1_p->n_annot && s_annot1_arr_p[i1_annot]->pos < n1) {
	if (s_annot1_arr_p[i1_annot]->label == '[') break;
	if (s_annot1_arr_p[i1_annot]->label == ']') {
	  push_stack(annot_stack, s_annot1_arr_p[i1_annot]);
	  have_push_features = 1;
	}
	i1_annot++;
      }
    }

    if (annot0_p) {
      /* also check for regions after alignment */
      while (i0_annot < annot0_p->n_annot && s_annot0_arr_p[i0_annot]->pos < n0) {
	if (s_annot0_arr_p[i0_annot]->label == '[') break;
	if (s_annot0_arr_p[i0_annot]->label == ']') {
	  push_stack(annot_stack, s_annot0_arr_p[i0_annot]);
	  have_push_features = 1;
	}
	i0_annot++;
      }
    }

    if (have_push_features) {
      display_push_features(annot_stack, annot_var_s, n_annot_var_s,
			    a_res->max0-1 + q_offset, *sp0, 
			    a_res->max1-1 + l_offset, *sp1,
			    sim_sym[*spa], &region0_p, &region1_p,
			    &(a_res->rst), n0, n1, pstat_void, DP_FULL_FMT);
    }
  }

  *spa = '\0';

#ifndef LCAL_CONS	/* have context around alignment */
  /*	now we have the middle, get the right end */
  if (!aln->llcntx_set) {
    ns = mins + lenc + aln->llen;	/* show an extra line? */
    ns -= (itmp = ns %aln->llen);	/* itmp = left over on last line */
    if (itmp>aln->llen/2) ns += aln->llen;  /* more than 1/2 , use another*/
    nd = ns - (mins+lenc);		/* this much extra */
  }
  else nd = aln->llcntx;

  if (nd > max(n0-a_res->max0,nn1-a_res->max1)) 
    nd = max(n0-a_res->max0,nn1-a_res->max1);
  
  if (aln->showall==1) {
    nd = max(n0-a_res->max0,nn1-a_res->max1);	/* reset for showall=1 */
    /* get right end */
    aancpy(seqc0+mins+lenc,(char *)aa0+a_res->max0,n0-a_res->max0,ppst);
    aancpy(seqc1+mins+lenc,(char *)aa1p+a_res->max1,nn1-a_res->max1,ppst);
    /* fill with blanks - this is required to use one 'nc' */
    memset(seqc0+mins+lenc+n0-a_res->max0,' ',nd-(n0-a_res->max0));
    memset(seqc1+mins+lenc+nn1-a_res->max1,' ',nd-(nn1-a_res->max1));
  }
  else {
    if ((nd-(n0-a_res->max0))>0) {
      aancpy(seqc0+mins+lenc,(char *)aa0+a_res->max0,(n0-a_res->max0),ppst);
      memset(seqc0+mins+lenc+n0-a_res->max0,' ',nd-(n0-a_res->max0));
    }
    else {
      aancpy(seqc0+mins+lenc,(char *)aa0+a_res->max0,nd,ppst);
    }

    if ((nd-(nn1-a_res->max1))>0) {
      aancpy(seqc1+mins+lenc,(char *)aa1p+a_res->max1,nn1-a_res->max1,ppst);
      memset(seqc1+mins+lenc+nn1-a_res->max1,' ',nd-(nn1-a_res->max1));
    }
    else {
      aancpy(seqc1+mins+lenc,(char *)aa1p+a_res->max1,nd,ppst);
    }
  }
  if (have_ann) {
    memset(seqc0a+mins+lenc,' ',nd);
    memset(seqc1a+mins+lenc,' ',nd);
    /*
      ntmp = nd-(n0-a_res->max0);
      if (ntmp > 0) memset(seqc0a+mins+lenc+n0-a_res->max0,' ',ntmp);
      ntmp = nd-(nn1-a_res->max1);
      if (ntmp > 0) memset(seqc1a+mins+lenc+nn1-a_res->max1,' ',ntmp);
    */
  }
#else
  nd = 0;
#endif

  /*  fprintf(stderr,"%d\n",mins+lenc+nd); */

  lenc = mins + lenc + nd;

  if (annot0_p || annot1_p) free_stack(annot_stack);

  return lenc;
}

void
calc_astruct(struct a_struct *aln_p, struct a_res_str *a_res_p) {

  /* we do not pay attention to aln_p->calc_last_set, because all the
     functions (calc_astruct, calc_cons_a, calc_code) use exactly the same
     assignment */

  aln_p->amin0 = a_res_p->min0;
  aln_p->amax0 = a_res_p->max0;
  aln_p->amin1 = a_res_p->min1;
  aln_p->amax1 = a_res_p->max1;
}

/* update_code has been modified to work more correctly with
   ggsearch/glsearch, which, because alignments can start with either
   insertions or deletions, can produce an initial code of "0=".  When
   that happens, it is ignored and no code is added. */

static void
update_code(char *al_str, int al_str_max, int op, int op_cnt, int show_code, int *start_code) {
  char *ori_code = "=-+*";
  char *cigar_code = "MDIM";

  char tmp_cnt[20];

  if (*start_code && op_cnt==0 && op == 0) {
    *start_code = 0;
    return;
  }

  *start_code = 0;

  if (show_code == SHOW_CODE_CIGAR) {
    sprintf(tmp_cnt,"%d%c",op_cnt,cigar_code[op]);
  }
  else {
    sprintf(tmp_cnt,"%c%d",ori_code[op],op_cnt);
  }
  strncat(al_str,tmp_cnt,al_str_max);
}

/* build an array of match/ins/del - length strings */
int calc_code(const unsigned char *aa0, int n0,
	      const unsigned char *aa1, int n1,
	      struct a_struct *aln,
	      struct a_res_str *a_res,
	      struct pstruct *ppst,
	      char *al_str, int al_str_n, 
	      const unsigned char *ann_arr, 
	      const unsigned char *aa0a,
	      const struct annot_str *annot0_p,
	      const unsigned char *aa1a,
	      const struct annot_str *annot1_p,
	      char *ann_code_s, int n_ann_code_s,
	      int *score_delta,
	      struct f_struct *f_str,
	      void *pstat_void,
	      int display_code)
{
  int i0, i1;
  int op, lenc;
  int p_op, op_cnt;
  const unsigned char *aa1p;
  char tmp_str[MAX_SSTR];
  char sp0, sp1;
  unsigned char *sq;
  int *rp;
  int have_ann=0;
  char ann_ch0, ann_ch1;
  char tmp_astr[MAX_STR];
  int sim_code, t_spa;
  char *sim_sym = aln_map_sym[MX_ACC];
  int aa0c, aa1c, itmp;
  int show_code, annot_fmt, start_code;

  /* variables for variant changes, regions */
  void *annot_stack;
  struct annot_entry **s_annot0_arr_p, *region0_p, tmp_annot0;
  struct annot_entry **s_annot1_arr_p, *region1_p, tmp_annot1;
  int  i0_annot, i1_annot, v_delta, v_tmp;
  int have_push_features;
  long q_offset, l_offset;

  *score_delta = 0;

  show_code = (display_code & SHOW_CODE_MASK);
  annot_fmt = 2;
  if (display_code & SHOW_ANNOT_FULL) {
    annot_fmt = 1;
  }
  
  if (aa0a != NULL && aa1a != NULL) { have_ann = 2;}
  else if (aa0a != NULL || aa1a != NULL) { have_ann = 1;}
  else {have_ann = 0;}

  if (ppst->ext_sq_set) {
    sq = ppst->sqx;
  }
  else {
    sq = ppst->sq;
  }

#ifndef TFASTA
  aa1p = aa1;
#else
  aa1p = f_str->aa1x;
#endif

  rp = a_res->res;
  lenc = aln->nident = aln->nmismatch = aln->nsim = aln->npos = aln->ngap_q = aln->ngap_l = aln->nfs = 0;

  i0 = a_res->min0;
  i1 = a_res->min1;

  q_offset = aln->q_offset;
  l_offset = aln->l_offset;

  op = 0;
  p_op = 0;
  op_cnt = 0;
  start_code = 1;

  v_delta = 0;
  i0_annot = i1_annot = 0;
  annot_stack = NULL;
  region0_p = region1_p = NULL;
  s_annot0_arr_p = s_annot1_arr_p = NULL;
  if (have_ann) {
    have_push_features = 0;
    if (annot0_p || annot1_p) annot_stack = init_stack(64,64);

    if (annot1_p && annot1_p->n_annot > 0) {
      s_annot1_arr_p = annot1_p->s_annot_arr_p;
      while (i1_annot < annot1_p->n_annot && s_annot1_arr_p[i1_annot]->pos < i1+l_offset) {
	if (s_annot1_arr_p[i1_annot]->label == '[') {
	  memcpy(&tmp_annot1,s_annot1_arr_p[i1_annot], sizeof(struct annot_entry));
	  tmp_annot1.pos = a_res->min1 + l_offset;
	  tmp_annot1.a_pos = a_res->min0 + q_offset;
	  region1_p = &tmp_annot1;
	  region1_p->score = region1_p->n_ident = region1_p->n_aln = 0;
	}
	else if (s_annot1_arr_p[i1_annot]->label == ']') {
	  region1_p = NULL;
	}
	i1_annot++;
      }
    }

    if (annot0_p && annot0_p->n_annot > 0) {
      s_annot0_arr_p = annot0_p->s_annot_arr_p;
      while (i0_annot < annot0_p->n_annot && s_annot0_arr_p[i0_annot]->pos < i0+q_offset) {
	if (s_annot0_arr_p[i0_annot]->label == '[') {
	  memcpy(&tmp_annot0,s_annot0_arr_p[i0_annot], sizeof(struct annot_entry));
	  tmp_annot0.pos = a_res->min0 + q_offset;
	  tmp_annot0.a_pos = a_res->min1 + l_offset;
	  region0_p = &tmp_annot0;
	  region0_p->score = region0_p->n_ident = region0_p->n_aln = 0;
	}
	else if (s_annot0_arr_p[i0_annot]->label == ']') {
	  region0_p = NULL;
	}
	i0_annot++;
      }
    }
  }

  while (i0 < a_res->max0 || i1 < a_res->max1) {
    /* match/mismatch (aligned residues */
    /* here, op is the "current" encoding, and *rp is the next one */
    if (op == 0 && *rp == 0) {

      aa0c = aa0[i0];
      aa1c = aa1p[i1];
      itmp = ppst->pam2[0][aa0c][aa1c];
      sp0 = sq[aa0c];
      sp1 = sq[aa1c];

      /* variant annot1_p annotations can cause substitution */
      if (s_annot1_arr_p) {
	if (i1+l_offset == s_annot1_arr_p[i1_annot]->pos) {
	  i1_annot = next_annot_match(&itmp, ppst->pam2[0][aa0c], l_offset+seq_pos(i1,aln->llrev,0),
				      q_offset+seq_pos(i0,aln->qlrev,0), &sp1, NULL, sq,
				      i1_annot, annot1_p->n_annot, s_annot1_arr_p,
				      NULL, annot_stack, &have_push_features, &v_delta,
				      &region1_p, &tmp_annot1, 0);
	}

	if (sq[aa1c] != sp1) {
	  t_spa = align_type(itmp, sp0, sp1, ppst->nt_align, NULL, ppst->pam_x_id_sim);
	  comment_var(q_offset+seq_pos(i0,aln->qlrev,0), sp0,
		      l_offset+seq_pos(i1,aln->llrev,0), sp1,
		      sq[aa1c], sim_sym[t_spa], NULL,
		      ann_code_s, n_ann_code_s, 1, annot_fmt);
	}
	if (region1_p) { region1_p->score += itmp;}
      }

      if (s_annot0_arr_p) {
	if (i0+q_offset == s_annot0_arr_p[i0_annot]->pos) {
	  i0_annot = next_annot_match(&itmp, ppst->pam2[0][aa1c], q_offset+seq_pos(i0,aln->qlrev,0),
				      l_offset+seq_pos(i1,aln->llrev,0), &sp0, NULL, sq,
				      i0_annot, annot0_p->n_annot, s_annot0_arr_p,
				      NULL, annot_stack, &have_push_features, &v_delta,
				      &region0_p, &tmp_annot0, 0);
	}

	/* check for sequence change from variant */
	if (sq[aa0c] != sp0) {
	  t_spa = align_type(itmp, sp0, sp1, ppst->nt_align, NULL, ppst->pam_x_id_sim);
	  comment_var(q_offset+seq_pos(i0,aln->qlrev,0), sp0,
		      l_offset+seq_pos(i1,aln->llrev,0), sp1,
		      sq[aa0c], sim_sym[t_spa], NULL,
		      ann_code_s, n_ann_code_s, 0, annot_fmt);
	}
	if (region0_p) { region0_p->score += itmp;}
      }

      sim_code = align_type(itmp, sp0, sp1, ppst->nt_align, aln, ppst->pam_x_id_sim);
      if (region0_p) {
	region0_p->n_aln++;
	if (sim_code == M_IDENT) {region0_p->n_ident++;}
      }

      if (region1_p) {
	region1_p->n_aln++;
	if (sim_code == M_IDENT) {region1_p->n_ident++;}
      }

      if (p_op == 0 || p_op==3) {	/* p_op -> previous op, 0->match, 1->del in seq0, 2->ins in seq0, 3->term match */
	if (sp0 != '*' && sp1 != '*') { /* neither is terminator */
	  if (p_op == 3) {
	    update_code(al_str,al_str_n-strlen(al_str),p_op,op_cnt, show_code, &start_code);
	    op_cnt = 1; p_op = 0;
	  }
	  else {op_cnt++;}
	}
	else {
	  update_code(al_str,al_str_n-strlen(al_str),p_op,op_cnt, show_code, &start_code);
	  op_cnt = 1; p_op = 3;
	}
      }
      else {
	update_code(al_str,al_str_n-strlen(al_str),p_op,op_cnt, show_code, &start_code);
	op_cnt = 1; p_op = 0;
      }

      /* update op to the next encoded position */
      op = *rp++;
      lenc++;
      
      /* check for an annotation */
      if (have_ann) {
	ann_ch0 = ann_ch1 = '\0';
	/* conventional annotations */
	if (have_ann == 2 && (ann_arr[aa0a[i0]] != ' ' || ann_arr[aa1a[i1]] != ' ')) {
	  ann_ch0 = ann_arr[aa0a[i0]];
	  if (ann_ch0 == ' ') ann_ch0 = 'X';
	  ann_ch1 = ann_arr[aa1a[i1]];
	  if (ann_ch1 == ' ') ann_ch1 = 'X';
	}
	else if (aa0a != NULL && ann_arr[aa0a[i0]]!=' ') {
	  ann_ch0 = ann_arr[aa0a[i0]];
	  ann_ch1 = 'X';
	}
	else if (aa1a != NULL && ann_arr[aa1a[i1]]!=' ') {
	  ann_ch0 = 'X';
	  ann_ch1 = ann_arr[aa1a[i1]];
	}

	/* ann_ch0 only works below because ann_ch0=='X' if ann_ch1 */
	if ( ann_ch0 && !(ann_ch1 == '['  || ann_ch1 == ']' || ann_ch0 == '[' || ann_ch0 == ']')) {
	  sprintf(tmp_astr, "|%c%c:%ld%c%c%ld%c",
		  ann_ch0,ann_ch1, q_offset+seq_pos(i0,aln->qlrev,0)+1,sp0,
		  sim_sym[sim_code], l_offset+seq_pos(i1,aln->llrev,0)+1,sp1);
	  SAFE_STRNCAT(ann_code_s, tmp_astr, n_ann_code_s);
	}
	  
	if ((s_annot1_arr_p || s_annot0_arr_p) && have_push_features) {
	  display_push_features(annot_stack, ann_code_s, n_ann_code_s,
				q_offset+seq_pos(i0,aln->qlrev,0), sp0,
				l_offset+seq_pos(i1,aln->llrev,0), sp1,
				sim_sym[sim_code], &region0_p, &region1_p,
				&(a_res->rst), n0, n1, pstat_void, annot_fmt);
	  have_push_features = 0;
	}
      }
      i0++; i1++;
    }
    else {	/* not in match run, in a gap */
      if (op == 0) {
	/* at a transition from match (previous) to indel (current) */
	if (region1_p) {region1_p->score += ppst->gdelval;}
	if (region0_p) {region0_p->score += ppst->gdelval;}
	op = *rp++;
      }
      if (region1_p) {region1_p->score += ppst->ggapval; region1_p->n_aln++;}
      if (region0_p) {region0_p->score += ppst->ggapval; region0_p->n_aln++;}

      if (op > 0) {
	if (p_op == 1) { op_cnt++;}
	else {
	  update_code(al_str,al_str_n - strlen(al_str),p_op,op_cnt,show_code, &start_code);
	  op_cnt = 1; p_op = 1;
	}
	if (s_annot1_arr_p) {
	  if (i1+l_offset == s_annot1_arr_p[i1_annot]->pos) {
	    i1_annot = next_annot_match(&itmp, ppst->pam2[0][aa0[i0]], l_offset+seq_pos(i1,aln->llrev,0),
					q_offset+seq_pos(i0,aln->qlrev,0), &sp1, NULL, sq, 
					i1_annot, annot1_p->n_annot, s_annot1_arr_p,
					NULL, annot_stack, &have_push_features, &v_delta,
					&region1_p, &tmp_annot1, ppst->ggapval+ppst->gdelval);

	    if (have_push_features) {
	      display_push_features(annot_stack, ann_code_s, n_ann_code_s,
				    q_offset+seq_pos(i0,aln->qlrev,0), sp0,
				    l_offset+seq_pos(i1,aln->llrev,0), sp1,
				    sim_sym[sim_code], &region0_p, &region1_p,
				    &(a_res->rst), n0, n1, pstat_void, annot_fmt);
	      have_push_features = 0;
	    }
	  }
	}
	op--; lenc++; i1++; aln->ngap_q++;
      }
      else {  /* (op < 0) */
	if (p_op == 2) {  op_cnt++;}
	else {
	  update_code(al_str,al_str_n - strlen(al_str),p_op,op_cnt,show_code, &start_code);
	  op_cnt = 1; p_op = 2;
	}
	if (s_annot0_arr_p) {
	  if (i0+q_offset == s_annot0_arr_p[i0_annot]->pos) {
	    i0_annot = next_annot_match(&itmp, ppst->pam2[0][aa1c], q_offset+seq_pos(i0,aln->qlrev,0),
					l_offset+seq_pos(i1,aln->llrev,0), &sp0, NULL, sq, 
					i0_annot, annot0_p->n_annot, s_annot0_arr_p,
					NULL, annot_stack, &have_push_features, &v_delta,
					&region0_p, &tmp_annot0,ppst->ggapval+ppst->gdelval);

	    if (have_push_features) {
	      display_push_features(annot_stack, ann_code_s, n_ann_code_s,
				    q_offset+seq_pos(i0,aln->qlrev,0), sp0,
				    l_offset+seq_pos(i1,aln->llrev,0), sp1,
				    sim_sym[sim_code], &region0_p, &region1_p,
				    &(a_res->rst), n0, n1, pstat_void, annot_fmt);
	      have_push_features = 0;
	    }
	  }
	}
	op++; lenc++; i0++; aln->ngap_l++;
      }
    }
  }

  /* all done, clean things up */

  update_code(al_str,al_str_n - strlen(al_str),p_op,op_cnt,show_code, &start_code);

  if (have_ann) {
    have_push_features = 0;
    if (s_annot1_arr_p) {
      /* also check for regions after alignment */
      while (i1_annot < annot1_p->n_annot && s_annot1_arr_p[i1_annot]->pos < n1) {
	if (s_annot1_arr_p[i1_annot]->label == '[') break;
	if (s_annot1_arr_p[i1_annot]->label == ']') {
	  push_stack(annot_stack, s_annot1_arr_p[i1_annot]);
	  have_push_features = 1;
	}
	i1_annot++;
      }
    }

    if (s_annot0_arr_p) {
      /* also check for regions after alignment */
      while (i0_annot < annot0_p->n_annot && s_annot0_arr_p[i0_annot]->pos < n0) {
	if (s_annot0_arr_p[i0_annot]->label == '[') break;
	if (s_annot0_arr_p[i0_annot]->label == ']') {
	  push_stack(annot_stack, s_annot0_arr_p[i0_annot]);
	  have_push_features = 1;
	}
	i0_annot++;
      }
    }

    if (have_push_features) {
      display_push_features(annot_stack, ann_code_s, n_ann_code_s,
			    q_offset+a_res->max0-1, sp0,
			    l_offset+a_res->max1-1, sp1,
			    sim_sym[sim_code], &region0_p, &region1_p,
			    &(a_res->rst), n0, n1, pstat_void, annot_fmt);
    }
  }

  if (annot0_p || annot1_p) free_stack(annot_stack);

  *score_delta = v_delta;
  return lenc;
}

int calc_id(const unsigned char *aa0, int n0,
	    const unsigned char *aa1, int n1,
	    struct a_struct *aln, 
	    struct a_res_str *a_res,
	    struct pstruct *ppst,
	    const struct annot_str *annot0_p,
	    const struct annot_str *annot1_p,
	    int *score_delta,
	    char *annot_var_s, int n_annot_var_s,
	    struct f_struct *f_str)
{
  int i0, i1, nn1;
  int op, lenc;
  char sp0, sp1;
  char tmp_str[MAX_SSTR];
  const unsigned char *aa1p;
  int *rp;
  unsigned char *sq;
  /* variables for variant changes */
  struct annot_entry **s_annot0_arr_p, *region0_p;
  struct annot_entry **s_annot1_arr_p, *region1_p;
  int  itmp, i0_annot, i1_annot, v_delta, v_tmp;
  long q_offset, l_offset;
  
  *score_delta = 0;
  annot_var_s[0] = '\0';

  if (ppst->ext_sq_set) { sq = ppst->sqx; }
  else { sq = ppst->sq; }

#ifndef TFASTA
  aa1p = aa1;
  nn1 = n1;
#else
  aa1p = f_str->aa1x;
  nn1 = f_str->n10;
#endif

  aln->amin0 = a_res->min0;
  aln->amax0 = a_res->max0;
  aln->amin1 = a_res->min1;
  aln->amax1 = a_res->max1;
  aln->calc_last_set = 1;

  q_offset = aln->q_offset;
  l_offset = aln->l_offset;

  rp = a_res->res;
  lenc = aln->nident = aln->nmismatch = aln->nsim = aln->ngap_q = aln->ngap_l = aln->nfs = op = 0;
  i0 = a_res->min0;
  i1 = a_res->min1;

  v_delta = 0;
  i0_annot = i1_annot = 0;

  if (annot1_p && annot1_p->n_annot > 0) s_annot1_arr_p = annot1_p->s_annot_arr_p;
  else s_annot1_arr_p = NULL;
  if (annot0_p && annot0_p->n_annot > 0) s_annot0_arr_p = annot0_p->s_annot_arr_p;
  else s_annot0_arr_p = NULL;

  while (i0 < a_res->max0 || i1 < a_res->max1) {
    if (op == 0 && *rp == 0) {
      /* op==0 -> we are in a match run, and current code is a match */
      op = *rp++;
      lenc++;

      itmp = ppst->pam2[0][aa0[i0]][aa1p[i1]];
      sp0 = sq[aa0[i0]];
      sp1 = sq[aa1p[i1]];

      if (s_annot1_arr_p && i1 + l_offset == s_annot1_arr_p[i1_annot]->pos) {
	i1_annot = next_annot_match(&itmp, ppst->pam2[0][aa0[i0]], l_offset+seq_pos(i1,aln->llrev,0),
				    q_offset+seq_pos(i0,aln->qlrev,0), &sp1, NULL, sq,
				    i1_annot, annot1_p->n_annot, s_annot1_arr_p,
				    NULL, NULL, NULL, &v_delta, NULL, NULL, itmp);

	/* must be out of the loop to capture the last value */
	if (sq[aa1p[i1]] != sp1) {
	  sprintf(tmp_str,"%c%d%c;",sq[aa1p[i1]],i1+1,sp1);
	  SAFE_STRNCAT(annot_var_s,tmp_str,n_annot_var_s);
	}
      }

      if (s_annot0_arr_p && i0 + q_offset == s_annot0_arr_p[i0_annot]->pos) {
	i0_annot = next_annot_match(&itmp, ppst->pam2[0][aa0[i0]], q_offset+seq_pos(i0,aln->qlrev,0),
				    l_offset+seq_pos(i1,aln->llrev,0), &sp0, NULL, sq,
				    i0_annot, annot0_p->n_annot, s_annot0_arr_p,
				    NULL, NULL, NULL, &v_delta, NULL, NULL, itmp);

	/* must be out of the loop to capture the last value */
	if (sq[aa0[i0]] != sp0) {
	  sprintf(tmp_str,"%c%d%c;",sq[aa0[i0]],i0+1,sp1);
	  SAFE_STRNCAT(annot_var_s,tmp_str,n_annot_var_s);
	}
      }

      /* updates nident, nsim, npos */
      align_type(itmp, sp0, sp1, ppst->nt_align, aln, ppst->pam_x_id_sim);

      i0++; i1++;
    }
    else {
      if (op==0) op = *rp++;
      if (op>0) {	/* inserts in seq0 */
	op--; lenc++; i1++; aln->ngap_q++;
      }
      else {		/* inserts in seq 1 */
	op++; lenc++; i0++; aln->ngap_l++;
      }
    }
  }
  *score_delta = v_delta;
  return lenc;
}
