#!/usr/bin/perl -w

# l_feat.pl gets a description line as its argument:
#
# l_feat.pl sp|P00502|GSTA1_RAT
#
# (1) read in the line
# (2) parse it to get the up_acc
# (3) return the tab delimited features
#

use strict;

use DBI;
use Getopt::Long;

my $host = "mysql-pearson";
my $db = "up_db";
my $user = "web_user";
my $pass = "uniprot";
my $port = 4124;

my $debug = 0;
my $tag = "";
my $verbose = 0;
GetOptions("debug=i" => \$debug,
	   "verbose=i" => \$verbose,
	   "host=s" => \$host,
	   "db=s" => \$db,
	   "user=s" => \$user,
	   "password=s" => \$pass,
	   "port=i" => \&port,
	  );

my $connect = "dbi:mysql(AutoCommit=>1,RaiseError=>1):database=$db";
$connect .= ";host=$host" if $host;
$connect .= ";port=$port" if $port;

my $dbh = DBI->connect($connect,
		       $user,
		       $pass
		      ) or die $DBI::errstr;

my %h_feature = (
#    ACT_SITE => '=',
#    MOD_RES => '*',
#    BINDING => '#',
#    SITE => '^',
#    METAL => '!',
#    VARIANT => 'V',
#    MUTAGEN => 'V',
    DOMAIN_N => '[',
    DOMAIN_C => ']',
    REPEAT_N => '[',
    REPEAT_C => ']',
    );

my $get_annots_id = $dbh->prepare('select * from features join annot using(acc) where id=? order by pos');
my $get_annots_acc = $dbh->prepare('select * from features join annot using(acc) where acc=? order by pos');
my $get_annots = $get_annots_id;

my ($tmp, $gi, $sdb, $acc, $id, $use_acc);

my $annot_line = shift @ARGV;

die "no sequence description\n" unless $annot_line;

$annot_line =~ s/^>//;

print ">$annot_line\n";
if ($annot_line =~ m/^gi\|/) {
  $use_acc = 1;
  ($tmp, $gi, $sdb, $acc, $id) = split(/\|/,$annot_line);
  }
  elsif ($annot_line =~ m/SP:(\w+)/) {
    $use_acc = 0;
    $sdb = 'sp';;
    $id = $1;
  }
  elsif ($annot_line =~ m/TR:(\w+)/) {
    $use_acc = 0;
    $sdb = 'tr';
    $id = $1;
  }
  else {
    $use_acc = 1;
    ($sdb, $acc, $id) = split(/\|/,$annot_line);
  }


  # remove version number
  unless ($use_acc) {
    $get_annots = $get_annots_id;
    $get_annots->execute($id);
  }
  else {
    $get_annots = $get_annots_acc;
    $acc =~ s/\.\d+$//;
    $get_annots->execute($acc);
  }

my ($pos, $label, $value, $comment);

my %annot = ();

while (($acc, $pos, $label, $value) = $get_annots->fetchrow_array()) {
  next unless ($label =~ m/^DOMAIN/ || $label =~ m/^REPEAT/);
  if ($label eq "REPEAT_N" || $label eq "DOMAIN_N") {
    %annot = ();
    $annot{beg} = $pos;
    $annot{descr} = $value;
  } elsif ($label eq "REPEAT_C" || $label eq "DOMAIN_C") {
    unless ($annot{beg}) {
      warn(" $acc $pos $label $value -- missing _N\n");
      next;
    } else { 
      $annot{end} = $pos;
      print join("\t",($annot{beg}, $annot{end}, $annot{descr})),"\n";
    }
  }
}


exit(0);
